% Water-filling 
%
% Initially written by:     Jan Vidar Rognsvg: Spring 2008    
%
% Last revised by:          Jan Vidar Rognsvg 	22.06.2008
%
% Input: Index of line to be water-filled, PBO of all lines in the system,
% capacity constraints for each line, (crosstalk coefficient matrix)
% Output: Bit rate of the line matching the system index, capacity at SSM,
% ratio of transmitted PSD vs fixed PSD, output signal-to-noise ratio. 
% (No constraint applied to rate in line 1.)
% 
% Ex: system=1;PBOx=[0 0 0];Rwtd=[0 15 15];XFE=0;
% [R,cap,PSDr,SNR]=waterfill(system,PBOx,Rwtd,XFE);
function [R,cap,PSDr,SNR]=waterfill(system,PBOx,Rwtd,XFE)
tic
defVDSLu;
if system ~= 1
    % Include next two lines to permit top filling (waterfillfast).
    %[R,cap,PBO,SNR]=topfillfast(system,PBOx,Rwtd);
    %return
    if (Rwtd(system)==0)
        R=0;
        PSDr=ones(1,NF)*inf;
        cap=0;
        SNR=0;
        return
    end
end
% Lines: Ignores extra lines if size(dist)>size(PBOx,2)
[Npsd,Hpsd,cap] = sys(PBOx,XFE);
N = Npsd(1:NF,system).';
H = Hpsd(1:NF,system).';
M = 10^(-MARG/10);       % Margin compared to shannons limit
SNR = M*H.^2./N;    % Channel signal-to-noise ratio with unity signal PSD

% Include next line to add fictitious penalty to the SNR.
% SNR=SNR/20;

Sipsds = Sipsd-PBOx(1,system);      % Adjust for PBO in own signal(always fixed)
% Waterfilling, limited by Pmax, PSDcon & Rwtd.
Pmax = 10.^(Sipsds/10)*(FH-FL)*10^6;% Total transmitted power. [mW]
PSDcon = 10^((Sipsds+3)/10);        % PSC constraint.
tol = 0.01*10^((Sipsds-Sipsd)/10);  % Tolerance to inaccuracy in power allocation.
SNRmean=10*log10(mean(SNR));        % To adjust initial waterline for extreme SNR in disturbing lines.
% To ensure convergence
if (system==1)
    q = 1;      
else            % Slow down convergence, want R close to Rwtd.
    q = 10^(0.1*(SNRmean-96)); 
    q = q*10^(5-2*Rwtd(system)/15);   % 10-> far off 100->slow
    q = max(q,10);
end
K = min(1./SNR)+Pmax/((FH-FL)*10^6*q);   % Initial waterline
P = max(K-1./SNR,0);    % Transmitted PSD after waterfilling
% Deduct transmitted PSD where no integer bit has been alllocated
b = log2(1+P.*SNR);     % Number of bits in each subchannel
b = floor(b/BWEstep)*BWEstep;   % BWEstep=1Hz, -> b=floor(b)
b(b>BWEM) = BWEM;       % Max bandwidth efficiency
Pb = (2.^b-1)./SNR;     % Transmitted PSD after bitloading
Ptot = sum(Pb)*dF*1e6;  % Total power for current bitloading
R = sum(b)*dF;          % Transmission rate in Mbps
iter=0;
% Gradual water-filling
% repeat while (Powerdiff>tolerance & b<BWEM in some carrier & R<Rwanted)
while abs(Pmax-Ptot)>tol && (min(b)~=BWEM||Ptot>Pmax+tol) && (R<Rwtd(system)||system==1) && (iter<100)
    K=K+(Pmax-Ptot)/((FH-FL)*10^6*q); % Increase K to reach Pmax
    iter=iter+1;
    P=max(K-1./SNR,0);
    Pc=min(P,PSDcon); %PSD constraint
    b=log2(1+Pc.*SNR);
    b=floor(b/BWEstep)*BWEstep;    
    b(b>BWEM)=BWEM;
    Pb=(2.^b-1)./SNR;
    Ptot=sum(Pb)*dF*1e6;
    R=sum(b)*dF;
    % PSD constraint: Stops when least value of P equals PSDcon
    if (min(P)==PSDcon && Ptot<Pmax)% Exclude next line to remove constraint
        Pmax=Ptot;
    end
end 
% In case of no convergence
if(iter==100)
    warning('Waterfill exceeds number of iterations!')
    Constraint=Rwtd(system);
    while(Ptot>Pmax)
        [tmp,i] = max(Pb);
        b(i) = b(i)-1;
        Pb=(2.^b-1)./SNR;
        Ptot=sum(Pb)*dF*1e6;
    end
    R=sum(b)*dF;
end
PbdB=10*log10(Pb);
PSDr=Sipsd-PbdB;   % Fixed spectrum compared to the transmitted spectrum
toc

% Just for  presentation:
P(1./SNR>P)=0; 
PdB=10*log10(P);
% Sets all values of P less than K equal to 0:
%P(P<1./SNR)=0;
% Comparing with fixed spectrum:
bf=log2(1+10^(Sipsd/10).*SNR);
bf=floor(bf/BWEstep)*BWEstep; % BWEstep=1Hz, -> b=floor(b)
bf(bf>BWEM)=BWEM;
bf(bf<BWEmin-1.e-6)=0;
Rf=sum(bf)*dF;     %Returnerer det samme som Rfixed. (Med normal NSR)

% Comparison in dBm:
figure(1)
SdB=ones(1,NF)*Sipsds;
%KdB=ones(1,NF)*10*log10(K);
%plot(F,KdB,':')
stairs(F,PdB,'-.')
hold;
stairs(F,PbdB,'k')
plot(F,SdB,'--')
NCRdB = 10*log10(1./SNR);
stairs(F,NCRdB,'r:')    
legend('Water-filling basis','Transmitted spectrum','Fixed Spectum','NCR')
axis([FL FH floor(min(NCRdB-2)) ceil(max(max(max(NCRdB,PdB),(SdB)))+2)])
if(system==1)
    title('Transmitted Power Spectral Density of Line 1','FontSize',12)
else
    title('Transmitted Power Spectral Density in chosen disturber','FontSize',12)
end
xlabel('Frequency, MHz','FontSize',12)
ylabel('PSD, dBm/Hz','FontSize',12)

% Bit loading:
figure(2)
stairs(F,b,'k')
hold;
stairs(F,bf,':')
legend('Water-filling','Fixed Spectrum')
axis([FL FH 0 ceil(max(b)+1)])
title('Bit Loading - Bandwidth Efficiency','FontSize',12)      
xlabel('Frequency, MHz','FontSize',12)
ylabel('Bandwidth Efficiency, b/s/Hz','FontSize',12)  

% Comparison in mW:
figure(3)
S=ones(1,NF)*10^(Sipsds/10);
%K=ones(1,NF)*K;
%plot(F,K,':')
stairs(F,P,'-.')
hold;
stairs(F,Pb,'k')
plot(F,S,'--')
stairs(F,1./SNR,'r:')    
legend('Water-filling basis','Transmitted spectrum','Fixed Spectum','NCR')
axis([FL FH 0 1.1*(max(max(1./SNR,P)))])
title('Transmitted Power Spectral Density in chosen system','FontSize',12)      
xlabel('Frequency, MHz','FontSize',12)
ylabel('PSD, mW/Hz','FontSize',12)